#include <QTRSensors.h>
#include <EEPROM.h>
//speed of OUR bluetooth adapter is 57600
//const uint8_t GLOBAL_DELAY = 2;
QTRSensors sensor;
const uint8_t SENSOR_BUF_SIZE = 8;
uint16_t SENSOR_BUF[SENSOR_BUF_SIZE];
int err_arr[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int err_arr_pos = -1;
const int B1_PIN = 10;
const int B2_PIN = 9;
const int LED_PIN = 12;
bool B1_PRESSED = false;
bool B2_PRESSED = false;
int L_MOTOR_EN_PIN = 3;
int L_MOTOR_DIR_PIN = 4;
int L_MOTOR_PWM_PIN = 5;
int R_MOTOR_EN_PIN = 8;
int R_MOTOR_DIR_PIN = 7;
int R_MOTOR_PWM_PIN = 6;

int BASE_SPEED = 200;
int MAX_SPEED = 255;
float kP = 0.155; // коэффициент пропорциональной обратной связи
float kD = 0.057; // коэффициент дифференциальной обратной связи
float kI = 0.0; // коэффициент интегральной обратной связи

void setup()
{
  Serial.begin(57600);
  setupHardware();
  motorsForward(0);
  loadCalibration();
  doCalibration();
  Serial.println("STARTING...");
  motorsForward(25, 25);
  delay(500);
  motorsBackward(25, 25);
  delay(500);
  motorsForward(0, 0);
}

void loop()
{
  //float prev_error = 0;
  float error; // Это ошибка положения
  error = bot_position();

  err_arr_pos = (err_arr_pos + 1) % 10;
  err_arr[err_arr_pos] = error;

  float K = kP * error;
  float D = (err_arr[err_arr_pos] - err_arr[(err_arr_pos + 11) % 10]) * kD;

  //  String ss;
  //  for (int i = 0; i < 10; i++)  {
  //    ss += String(err_arr[i]) + " ";
  //  }
  //  Serial.println(ss);

  int err_sum = 0;
  for (int i = 0; i < 10; i++) err_sum += err_arr[i];
  float I = err_sum / 10 * kI;
  float correction =  K + I + D;

  int leftSpeed = BASE_SPEED * (1 - correction);
  int rightSpeed = BASE_SPEED * (1 + correction);
  int ls = leftSpeed;
  int rs = rightSpeed;

//    if (ls > 255) {
//      rs = rs - (ls - 255);
//      ls = 255;
//    }
//  
//    if (rs > 255) {
//      ls = ls - (rs - 255);
//      rs = 255;
//    }

  leftSpeed = constrain(ls, 0, MAX_SPEED);
  rightSpeed = constrain(rs, 0, MAX_SPEED);
  //String s = " LS:" + String(leftSpeed) + " RS:" + String(rightSpeed)  + " D:" + String(D) + " ls:" + String(ls) + " rs:" + String(rs);
  //Serial.println(s);
  motorsForward(leftSpeed, rightSpeed);
  String s = " POS:" + String(error) + " CORR:" + String(correction) + " L:" + String(leftSpeed) + " R:" + String(rightSpeed);
  //Serial.println(s);
  //delay(1);
}


float bot_position() {
  int pos1 = (sensor.readLineBlack(SENSOR_BUF) - 3500);
  //  Serial.println(String(pos1));
  float pos2 = pos1 / 1000.00;
  //Serial.println(String(pos2));
  return pos2;
}



void doCalibration() {
  Serial.println("PRESS B2 FOR CALIBRATION");
  for (int i = 0; i < 800; i++)
  {
    if (i % 50 == 0)
      digitalWrite(LED_PIN, HIGH);
    if (digitalRead(B2_PIN))
    {
      digitalWrite(LED_PIN, HIGH);
      for (uint16_t i = 0; i < 255; i++)  {
        Serial.println("CALIBRATION...");
        sensor.calibrate();
      }
      digitalWrite(LED_PIN, LOW);
      saveCalibration();
      break;
    }
    if (i % 100 == 0)
      digitalWrite(LED_PIN, LOW);
    delay(5);
  }

}

void saveCalibration() {
  Serial.print("SAVING CALIBRATION...");
  for (int i = 0; i < SENSOR_BUF_SIZE; i++) {
    EEPROM.put(i * 2, sensor.calibrationOn.minimum[i]);
    EEPROM.put(i * 2 + SENSOR_BUF_SIZE * 2, sensor.calibrationOn.maximum[i]);
  }
  Serial.println("DONE");
}

void loadCalibration() {
  Serial.print("LOADING CALIBRATION...");
  sensor.calibrate();  //init calibration array first
  for (int i = 0; i < SENSOR_BUF_SIZE; i++) {
    EEPROM.get(i * 2, sensor.calibrationOn.minimum[i]);
    EEPROM.get(i * 2 + SENSOR_BUF_SIZE * 2, sensor.calibrationOn.maximum[i]);
  }
  Serial.println("DONE");
}


void setupHardware() {
  Serial.print("SETUP HARDWARE...");
  pinMode(B1_PIN, INPUT);
  pinMode(B2_PIN, INPUT);
  pinMode(LED_PIN, OUTPUT);
  pinMode(L_MOTOR_DIR_PIN, OUTPUT);
  pinMode(R_MOTOR_DIR_PIN, OUTPUT);
  pinMode(L_MOTOR_EN_PIN, OUTPUT);
  pinMode(R_MOTOR_EN_PIN, OUTPUT);

  // configure the sensors
  sensor.setTypeAnalog();
  sensor.setSensorPins((const uint8_t[]) {
    A0, A1, A2, A3, A4, A5, A6, A7
  }, SENSOR_BUF_SIZE);
  Serial.println("DONE");
}

void motorsForward(int leftSpeed, int rightSpeed) {
  if (leftSpeed == 0) digitalWrite(L_MOTOR_EN_PIN, HIGH); else digitalWrite(L_MOTOR_EN_PIN, LOW);
  if (rightSpeed == 0) digitalWrite(R_MOTOR_EN_PIN, HIGH); else digitalWrite(R_MOTOR_EN_PIN, LOW);

  digitalWrite(L_MOTOR_DIR_PIN, HIGH);
  digitalWrite(R_MOTOR_DIR_PIN, LOW);
  analogWrite(L_MOTOR_PWM_PIN, leftSpeed);
  analogWrite(R_MOTOR_PWM_PIN, rightSpeed);
}

void motorsForward(int motorsSpeed) {
  motorsForward(motorsSpeed, motorsSpeed);
}

void motorsBackward(int leftSpeed, int rightSpeed) {
  digitalWrite(L_MOTOR_DIR_PIN, LOW);
  digitalWrite(R_MOTOR_DIR_PIN, HIGH);
  analogWrite(L_MOTOR_PWM_PIN, leftSpeed);
  analogWrite(R_MOTOR_PWM_PIN, rightSpeed);
}

void motorsBackward(int motorsSpeed) {
  motorsBackward(motorsSpeed, motorsSpeed);
}
